use std::{
    error::Error, str::FromStr, sync::Arc, net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, SocketAddrV4, SocketAddrV6}
};

use ignore_result::Ignore;
use log::info;

use tokio::{
    io::{AsyncBufReadExt, AsyncReadExt, AsyncWriteExt, BufReader}, 
    net::{TcpListener, TcpStream}
};
use tokio_io_timeout::TimeoutStream;
use tokio_rustls::TlsAcceptor;

use super::{
    tls::create_server_config,
    config::{
        Config,
        SiteConfig,
        IpForwarding
    }
};

pub struct FlowgateServer {
    config: Arc<Config>,
}

struct Connection {
    stream: BufReader<TcpStream>, 
    config: SiteConfig,
    keep_alive: bool, 
    host: String,
}

impl FlowgateServer {
    pub fn new(config: Arc<Config>) -> Self {
        FlowgateServer { config }
    }

    pub async fn start(self) {
        let local_self = Arc::new(self);

        tokio::spawn({ 
            let local_self = local_self.clone();
            async move { local_self.run_http().await.ignore(); } 
        });

        tokio::spawn({ 
            let local_self = local_self.clone();
            async move { local_self.run_https().await.ignore(); } 
        });
    }

    pub async fn run_http(self: Arc<Self>) -> Result<(), Box<dyn Error>> {
        let listener = TcpListener::bind(&self.config.http_host).await?;

        info!("HTTP server runned on {}", &self.config.http_host);

        loop {
            let Ok((stream, addr)) = listener.accept().await else { break }; 

            let local_self = self.clone();

            tokio::spawn(async move {
                let mut stream = TimeoutStream::new(stream);

                stream.set_write_timeout(Some(local_self.config.connection_timeout));
                stream.set_read_timeout(Some(local_self.config.connection_timeout));

                let mut stream = Box::pin(stream);

                local_self.accept_stream(
                    &mut stream,
                    addr,
                    false
                ).await;
            });
        }

        Ok(())
    }

    pub async fn run_https(self: Arc<Self>) -> Result<(), Box<dyn Error>> {
        let listener = TcpListener::bind(&self.config.https_host).await?;
        let acceptor = TlsAcceptor::from(Arc::new(create_server_config(self.config.clone()).await));

        info!("HTTPS server runned on {}", &self.config.https_host);

        loop {
            let Ok((stream, addr)) = listener.accept().await else { break }; 

            let local_self = self.clone();
            let acceptor = acceptor.clone();

            tokio::spawn(async move {
                let mut stream = TimeoutStream::new(stream);

                stream.set_write_timeout(Some(local_self.config.connection_timeout));
                stream.set_read_timeout(Some(local_self.config.connection_timeout));

                let Ok(mut stream) = acceptor.accept(Box::pin(stream)).await else { return };

                local_self.accept_stream(
                    &mut stream,
                    addr,
                    true
                ).await;
            });
        }

        Ok(())
    }

    async fn accept_stream(
        self: Arc<Self>, 
        stream: &mut (impl AsyncReadExt + AsyncWriteExt + Unpin), 
        addr: SocketAddr,
        https: bool
    ) -> Option<()> {
        let mut conn = self.clone().read_request(stream, addr, https, None).await?;

        if conn.keep_alive && conn.config.enable_keep_alive {
            loop {
                if !conn.config.support_keep_alive {
                    conn.stream.shutdown().await.ignore();
                    conn.stream = BufReader::new(conn.config.connect().await?);
                }
                conn = self.clone().read_request(stream, addr, https, Some(conn)).await?;
            }
        }

        conn.stream.shutdown().await.ignore();
        stream.shutdown().await.ok()?;

        Some(())
    }

    async fn read_request(
        self: Arc<Self>, 
        stream: &mut (impl AsyncReadExt + AsyncWriteExt + Unpin), 
        addr: SocketAddr,
        https: bool,
        conn: Option<Connection>
    ) -> Option<Connection> {
        let mut addr = addr;

        let mut stream = BufReader::new(stream);

        match &self.config.incoming_ip_forwarding {
            IpForwarding::Simple => {
                let mut header = Vec::new();
                stream.read_until(b'\n', &mut header).await.ok()?;
                header.truncate(header.len()-1);

                addr = SocketAddr::from_str(&String::from_utf8(header).ok()?).ok()?;
            },
            IpForwarding::Modern => {
                let mut header = [0];
                stream.read(&mut header).await.ok()?;

                addr = match header[0] {
                    0x01 => {
                        let mut octets = [0; 4];
                        stream.read(&mut octets).await.ok()?;

                        let mut port = [0; 2];
                        stream.read(&mut port).await.ok()?;
                        let port = u16::from_be_bytes(port);

                        SocketAddr::V4(SocketAddrV4::new(Ipv4Addr::from(octets), port))
                    }, 0x02 => {
                        let mut octets = [0; 16];
                        stream.read(&mut octets).await.ok()?;

                        let mut port = [0; 2];
                        stream.read(&mut port).await.ok()?;
                        let port = u16::from_be_bytes(port);

                        SocketAddr::V6(SocketAddrV6::new(Ipv6Addr::from(octets), port, 0, 0))
                    }, _ => { return None },
                };
            },
            _ => {}
        }

        let mut raw_status = read_until(&mut stream, b"\r\n").await?;
        raw_status.truncate(raw_status.len() - 2);
        let status =  String::from_utf8(raw_status.clone()).ok()?;
        let status = status.split(" ").collect::<Vec<&str>>();

        let mut content_length = 0;
        let mut host = None;
        let mut is_chunked = false;
        let mut keep_alive = false;

        let mut headers = Vec::new();

        loop {
            let mut header = read_until(&mut stream, b"\r\n").await?;
            header.truncate(header.len() - 2);

            if header.is_empty() {
                break;
            }

            let header = String::from_utf8(header).ok()?;
            let (key, value) = header.split_once(": ")?;

            headers.push((key.to_string(), value.to_string()));
            
            match key.to_lowercase().as_str() {
                "transfer-encoding" => {
                    if value.contains("chunked") {
                        is_chunked = true;
                    }
                },
                "host" => {
                    host = Some(value.to_string());
                },
                "connection" => {
                    keep_alive = value.to_lowercase().contains("keep-alive");
                },
                "content-length" => {
                    content_length = value.parse::<usize>().ok()?;
                },
                _ => {
                    if let IpForwarding::Header(header) = &self.config.incoming_ip_forwarding {
                        if key.to_lowercase() == header.to_lowercase() {
                            addr = SocketAddr::from_str(value).ok()?;
                        }
                    }
                },
            }
        }

        let mut conn: Connection = if conn.is_none() {
            let host = host?;
            let site = self.config.get_site(&host)?.clone();

            Connection {
                stream: BufReader::new(site.connect().await?),
                config: site,
                keep_alive,
                host
            }
        } else {
            conn?
        };

        let mut request = Vec::new();

        match &conn.config.ip_forwarding {
            IpForwarding::Simple => {
                request.append(&mut addr.to_string().as_bytes().to_vec());
                request.push(b'\n');
            },
            IpForwarding::Modern => {
                match addr.ip() {
                    IpAddr::V4(ip) => {
                        request.push(0x01);
                        request.append(&mut ip.octets().to_vec());
                    }, IpAddr::V6(ip) => {
                        request.push(0x02);
                        request.append(&mut ip.octets().to_vec());
                    }
                }
                request.append(&mut addr.port().to_be_bytes().to_vec());
            },
            _ => {}
        }

        request.append(&mut raw_status.clone());

        for (key, value) in headers {
            let mut value = value.to_string();

            match key.to_lowercase().as_str() {
                "host" => {
                    if let Some(replace_host) = conn.config.replace_host.clone() {
                        value = replace_host;
                    }
                },
                _ => {}
            }

            if let IpForwarding::Header(header) = &conn.config.ip_forwarding {
                if key.to_lowercase() == header.to_lowercase() {
                    continue;
                }
            }

            request.append(&mut key.as_bytes().to_vec());
            request.append(&mut b": ".to_vec());
            request.append(&mut value.as_bytes().to_vec());
            request.append(&mut b"\r\n".to_vec());
        }
        
        if let IpForwarding::Header(header) = &conn.config.ip_forwarding {
            request.append(&mut header.as_bytes().to_vec());
            request.append(&mut b": ".to_vec());
            request.append(&mut addr.to_string().as_bytes().to_vec());
            request.append(&mut b"\r\n".to_vec());
        }

        request.append(&mut b"\r\n".to_vec());

        conn.stream.write_all(&request).await.ok()?;

        if content_length > 0 {
            let buffer = stream.buffer();

            conn.stream.write_all(buffer).await.ok()?;
            stream.consume(buffer.len());

            let mut read = 0usize;
            let mut buf = vec![0; 4096];

            while let Ok(size) = stream.get_mut().read(&mut buf).await {
                if size == 0 { break }
                read += size;
                buf.truncate(size);
                conn.stream.write_all(&buf).await.ok()?;
                buf = vec![0; 4096];
                if read >= content_length { break }
            }
        } else if is_chunked {
            transfer_chunked(&mut stream, &mut conn.stream).await?;
        }

        if conn.config.support_keep_alive {
            let raw_status = read_until(&mut conn.stream, b"\r\n").await?;
            
            stream.write_all(&raw_status).await.ok()?;

            let mut content_length = 0;
            let mut is_chunked = false;

            loop {
                let mut header = read_until(&mut conn.stream, b"\r\n").await?;

                if header.len() == 2 {
                    break;
                }

                stream.write_all(&header).await.ok()?;

                header.truncate(header.len() - 2);

                let header = String::from_utf8(header).ok()?;
                let (key, value) = header.split_once(": ")?;
                
                match key.to_lowercase().as_str() {
                    "transfer-encoding" => {
                        if value.contains("chunked") {
                            is_chunked = true;
                        }
                    },
                    "content-length" => {
                        content_length = value.parse::<usize>().ok()?;
                    },
                    _ => {}
                }
            }

            if content_length > 0 {
                let buffer = conn.stream.buffer();

                stream.write_all(buffer).await.ok()?;
                stream.consume(buffer.len());

                let mut read = 0usize;
                let mut buf = vec![0; 4096];

                while let Ok(size) = conn.stream.get_mut().read(&mut buf).await {
                    if size == 0 { break }
                    read += size;
                    buf.truncate(size);
                    stream.write_all(&buf).await.ok()?;
                    buf = vec![0; 4096];
                    if read >= content_length { break }
                }
            } else if is_chunked {
                transfer_chunked(&mut conn.stream, &mut stream).await?;
            }
        } else {
            let buffer = conn.stream.buffer();

            stream.write_all(buffer).await.ok()?;
            stream.consume(buffer.len());

            let mut buf = vec![0;1024];
            while let Ok(n) = conn.stream.get_mut().read(&mut buf).await {
                if n == 0 { break }
                buf.truncate(n);
                stream.write_all(&buf).await.ok()?;
                buf = vec![0;1024];
            }
        }

        info!("{addr} > {} {}://{}{}", status[0], if https { "https" } else { "http" }, conn.host, status[1]);

        Some(conn)
    }
}

async fn read_until(stream: &mut (impl AsyncBufReadExt + Unpin), delimiter: &[u8]) -> Option<Vec<u8>> {
    let mut data = Vec::new();

    let last_byte = *delimiter.last()?;

    loop {
        let mut buf = Vec::new();
        stream.read_until(last_byte, &mut buf).await.ok()?;
        data.append(&mut buf);
        if data.ends_with(delimiter) {
            break;
        }
    }

    Some(data)
}

async fn transfer_chunked(src: &mut (impl AsyncBufReadExt + Unpin), dest: &mut (impl AsyncWriteExt + Unpin)) -> Option<()> {
    loop {
        let mut length = read_until(src, b"\r\n").await?;
        dest.write_all(&length).await.ok()?;
        length.truncate(length.len()-2);
        let length = String::from_utf8(length).ok()?;
        let length = usize::from_str_radix(length.as_str(), 16).ok()?;

        let mut data = vec![0u8; length+2];
        src.read_exact(&mut data).await.ok()?;
        dest.write_all(&data).await.ok()?;

        if length == 0 {
            break;
        }
    }

    Some(())
}